/*
    LineBufferedOutputStream.java - buffering based on new lines
    Copyright (C) 1999 Mark Benvenuto <mcb54@columbia.edu>

    This code was derived from 
	 BufferedOutputStream.java -- Buffer output into large blocks before writing

	 Copyright (c) 1998 Free Software Foundation, Inc.
	 Written by Aaron M. Renn (arenn@urbanophile.com)
    It is available from http://www.classpath.org/cvs/java/io

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package gnu.inet.tftp;

//import java.lang.*;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
/**
LineBufferedOutputStream

A class to buffer output to a stream based on lines and a fixed buffer.
This class is based on code written by the GNU Classpath project. It is
taken from BufferedOutputStream.java by Aaron M. Renn (arenn@urbanophile.com)
as part of the java.io package and it is licensed under GPL.
Note: There is no reason it has be in this namespace but it is for 
convience. 

@author Mark Benvenuto
@version 0.1

*/
public class LineBufferedOutputStream extends FilterOutputStream
{
	protected byte buf[];
	protected int count;
	private static final int DEFAULT_BUFFER_SIZE = 512;
	//protected String lineSeparator;
	
	/**
		Create a new LineBufferedOutputStream

		@param out OutputStream to buffer
	*/
	public LineBufferedOutputStream(OutputStream out)
	{
		this(out, DEFAULT_BUFFER_SIZE);
	}

	/**
		Create a new LineBufferedOutputStream

		@param out OutputStream to buffer
		@param size Size of internal buffer to hold data
	*/
	public LineBufferedOutputStream(OutputStream out, int size)
	{
		super(out);

		buf = new byte[size];
		//lineSeparator = System.getProperty("line.separator");
	}

	/**
		Flushes the buffer line by line
		and then does a complete flush for the remaining data
	*/
	public synchronized void
	flush() throws IOException
	{

		//for(int i = 0; i < count; i++)
		        //System.out.println("i["+i+"]:"+(int)buf[i]);
		for(int i = 0; i < count; i++)
			if( buf[i] == '\n') {
				//System.out.println(i);
				out.write( buf, 0, i + 1);
				System.arraycopy(buf, i + 1, buf, 0, count - (i + 1) );
				count -= (i + 1);
				i = 0;
				continue;
			}
		if (count == 0)
			return;

		out.write(buf, 0, count);
		count = 0;
	}

	/**
		Writes byte b to the internal buffer and flushes it if
		it is too big

		@param b a byte to write
	*/
	public synchronized void write(int b) throws IOException
	{
		buf[count] = (byte)(b & 0xFF);

		++count;
		if (count == buf.length)
			flush();
	}

	/**
		Writes bytes buf to the internal buffer and flushes it if
		it is too big

		@param buf bytes to write
		@param offset from where to start write
		@param len length of byte to write
	*/
	public synchronized void write(byte[] buf, int offset, int len) throws IOException
	{
		// Buffer can hold everything
		if (len < (this.buf.length - count))
		{
			System.arraycopy(buf, offset, this.buf, count, len);
			count += len;

			if (count == buf.length)
				flush();
		}
		else
		{
			flush();

			// Loop and write out full buffer chunks. As an optimization, 
			// don't buffer these, just write them
			int i = 0;
			if ((len / this.buf.length) != 0)
				for (i = 0; i < (len / this.buf.length); i++)
					out.write(buf, offset + (i * this.buf.length), this.buf.length);

			// Buffer the remaining bytes
			if ((len % buf.length) != 0)
			{
				System.arraycopy(buf, offset + (i * this.buf.length), this.buf, count, 
					len - (i * this.buf.length));
				count += (len - (i * this.buf.length));
			}
		}
	}
}
