/*
    TftpConnection.java - a TFTP Protocol Version 2 Client Connection Manager
    Copyright (C) 1999 Mark Benvenuto <mcb54@columbia.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package gnu.inet.tftp;

//import java.lang.*;
import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.StringTokenizer;

/**
TftpConnection

A class to manage either a read or write tftp request

@author Mark Benvenuto
@version 0.7
@see Tftp
@see TftpReadConnection 
@see TftpWriteConnection 

*/
public class TftpConnection implements Runnable
{
	InetAddress cia;
	DatagramSocket conn;
	int localport, remoteport;
	String remotefile, localfile;
	int error, debug, maxPacketSize = 516;
	int TransferType;
	static final short TFTP_RRQ = 1;
	static final short TFTP_WRQ = 2;
	static final short TFTP_DATA = 3;
	static final short TFTP_ACK = 4;
	static final short TFTP_ERROR = 5;
	static final String TFTP_netascii = "netascii";
	static final String TFTP_octet = "octet";
	static final String TFTP_mail = "mail";

	/**
		Creates a new instance of TftpConnection

		@param ia The internet address of the remote host
		@param remoteFilename The filename of the remote file
		@param localFilename The filename of the local file
		@param rtype The request type, 1 for read, 2 for write
		@param dbstate A flag to enable debug/verbose support, 0 for no debuging
		@see java.net.InternetAddress
	*/
	TftpConnection(InetAddress ia, String remoteFilename, String localFilename, int rtype, int dstate)
	{
		debug = dstate;
		cia = ia;
		remotefile = remoteFilename;
		localfile = localFilename;
		TransferType = rtype;

		//Create Socket
		try{
			conn = new DatagramSocket();
		}catch( SocketException e)
		{
			System.err.println("Datagram Socket Creation Failed: " + e);
			conn = null;
			return;
		}
		
		localport = conn.getLocalPort();

	}
	/**
		Processes the TftpConnection
	*/
	public void run()
	{
		if(conn == null)
			return;
		if(TransferType == 1)  //Read Request - receive file
		{
			//Read the File
			if(debug > 0)
				System.out.println("Reading the file: " + localfile);

			remoteport = 69; //Default TFTP port
	
			if(debug > 0)
				System.out.println("Created Socket: " + conn);

			//Send Read Request
			try{
				conn.send(BuildRequestPacket((short)1, remotefile, TFTP_netascii ));
			}catch( IOException e)
			{
				System.err.println("Error on Sending RRQ Packet: " + e);
			}

			if(debug > 0)
				System.out.println("Sent RRQ Packet");

			//Hande Data
			File outFile = new File(localfile);
			try{
				FileOutputStream fstream = new FileOutputStream( localfile );

				Date sd = new Date();
				TftpReadConnection tftpRConn = new TftpReadConnection( conn, cia, fstream, debug );
				tftpRConn.processConnection();
				Date ed = new Date();
				fstream.close();

				long filesize = new File( localfile ).length();
				System.out.println("Transfer successful: " + generateTransferStatistics( sd, ed, filesize) );
			}catch(IOException IOe)
			{
				System.err.println("FileOutputStream Creation failed: " + IOe);
				return;
			}
		}
		else if(TransferType == 2) //Write Request - send the file
		{
			//Send the File
			if(debug > 0)
				System.out.println("Writing the file: " + localfile);

			remoteport = 69; //Default TFTP port
	
			//Hande Data
			//Check here so that we don't initiate a write even though
			//we might not have the file
			File inFile = new File(localfile);
			if(inFile.exists() == false)
			{
				System.out.println("File does not exist");
				return;
			}

			if(debug > 0)
				System.out.println("Created Socket: " + conn);

			//Send Write Request
			try{
				conn.send(BuildRequestPacket((short)2, remotefile, TFTP_netascii ));
			}catch( IOException e)
			{
				System.err.println("Error on Sending WRQ Packet: " + e);
			}

			if(debug > 0)
				System.out.println("Sent WRQ Packet");

			try{
				FileInputStream fstream = new FileInputStream( localfile );
				//BufferedInputStream bfstream = new BufferedInputStream( fstream, 512 );

				Date sd = new Date();
				TftpWriteConnection tftpWConn = new TftpWriteConnection( conn, cia, fstream, debug );
				tftpWConn.processConnection();
				Date ed = new Date();
				fstream.close();


				long filesize = new File( localfile ).length();
				System.out.println("Transfer successful: " + generateTransferStatistics( sd, ed, filesize) );
			}catch(IOException IOe)
			{
				System.err.println("FileInputStream Creation failed: " + IOe);
				return;
			}
		}
	}
	/**
		Sets the socket timeout for the connection socket 
		which applies also to TftpReadConnection and TftpWriteConnection.

		@param time The new socket timeout in milliseconds
		@see TftpReadConnection 
		@see TftpWriteConnection 
	*/
	void setSocketTimeout(int time)
	{	
		try{
			conn.setSoTimeout(time);
		}catch( SocketException e)
		{
			System.err.println("Error on setting socket timeout: " + e);
		}
	}	
	/**
		Gets the socket timeout for the connection socket 
		which applies also to TftpReadConnection and TftpWriteConnection.

		@return The socket timeout in milliseconds
		@see TftpReadConnection 
		@see TftpWriteConnection 
	*/
	int getSocketTimeout() throws SocketException
	{	
		return conn.getSoTimeout();
	}	
	/**
		Creates a proper read/write request packet
		
		@param Opcode Type of request, 1 for read, 2 for write
		@param Filename The filename to pass the other computer
		@param mode The mode to use, usually netascii
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildRequestPacket(short Opcode, String Filename, String Mode)
	{
		byte data[];
		int pos;
		data = new byte[4 + Filename.length() + Mode.length()];
	

		data[0] = 0;
		data[1] = (byte)(0xF & Opcode);
		pos = 2;

		//Store Filename
	        byte bytes[] = Filename.getBytes();
        	System.arraycopy(bytes, 0, data, pos, bytes.length);
	        data[pos + bytes.length] = 0;
		pos += bytes.length + 1;

		//Store Mode
	        bytes = Mode.getBytes();
        	System.arraycopy(bytes, 0, data, pos, bytes.length);
	        data[pos + bytes.length] = 0;
		
		return new DatagramPacket(data, data.length, cia, remoteport);
	}
	/**
		Creates a proper DATA packet
		
		@param block The block number
		@param bytes The data stored in an array of lenght len
		@param len The length of bytes
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildDATAPacket(int block, byte[] bytes, int len)
	{
		byte data[];
		data = new byte[4 + len];

		data[0] = 0;
		data[1] = 3;
		
		data[2] = (byte)(0xFF & (block >> 8));
		data[3] = (byte)(0xFF & block);

		//Store Data
        	System.arraycopy(bytes, 0, data, 4, len);
		
		return new DatagramPacket(data, data.length, cia, remoteport);	
	}
	/**
		Creates a proper ACKnowledge packet
		
		@param block The block number
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildACKPacket(int block)
	{
		byte data[];
		data = new byte[4];

		data[0] = 0;
		data[1] = 4;
		
		data[2] = (byte)(0xFF & (block >> 8));
		data[3] = (byte)(0xFF & block);

		return new DatagramPacket(data, data.length, cia, remoteport);	
	}
	/*
		Used to read a string from an arrary of bytes, like a packet
	
		@param rs Bytes from a packet to read for string
		@param off Offset is the position to start reading from
		@return A String from the packet, could be zero length
	*/
	protected String readString(byte[] rs, int off)
	{
		String s = new String();
		off--;
		while(rs[++off]!=(byte)'\0')
			s += String.valueOf((char)rs[off]);
		return s;
	}
	/**
		Used to generate statistics about the number of bytes 
		transfered, the time, and the bytes/second.

		@param sd Starting Date
		@param ed Ending Date
		@param len Number of bytes transfered

		@return A String containing information that describes 
			the recent operation.
	*/
	String generateTransferStatistics( Date sd, Date ed, long len )
	{
		double time = (ed.getTime() - sd.getTime()) / 1000;
		return ( len + " bytes transfered in " + time + " : " + (len/time) + "(bytes/sec)" );
	}	
}
