/*
    TftpLogStream.java - a log stream capable of outputing to multiple streams
    Copyright (C) 1999 Mark Benvenuto <mcb54@columbia.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package gnu.inet.tftp;

//import java.lang.*;
import java.io.BufferedOutputStream;
//import java.io.BufferedWriter;
import java.io.IOException;
//import java.io.PrintWriter;
import java.io.PrintStream;
import java.io.OutputStream;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.TimeZone;
/**
TftpLogStream is a class designed to simultaneously output
to stdout and another PrintStream and pad each output with the
date and time.

Should be buffered by BufferedOutputStream or the included
LineBufferedOutputStream to work properly as the date prepending
will only work within the write(,,) functions and not write(int).

@author Mark Benvenuto
@version 0.1
@see LineBufferedOutputStream
*/
public class TftpLogStream extends OutputStream
{
	private boolean useStdout = true;
	private PrintStream logStdout = null;
	private boolean usePrimary = false;
	private PrintStream logPrimary = null;
	private boolean useDate = false;
	private Date logDate = null;
	private SimpleDateFormat logDateFormat = null;

	/**
		Constructs a TftpLogStream. Initializes date handling and by default 
		does not use it. By default it logs to Stdout.
	*/
	public TftpLogStream()
	{
		intializeDateHandling();
		logStdout = System.out;
	}

	/**
		Writes a byte to Stdout and Primary if needed. Note: This does not use
		the date buffering. If you want to use this class properly you <B>Must</B>
		buffer this OutputStream with BufferedOuputStream.

		@param b Byte to write to outputstreams
	*/
	public void write( int b ) throws IOException
	{
		//logStdout.println("Using write 1");
		if( useStdout == true)
		{
			logStdout.write(b);
		}
		if( usePrimary == true)
		{
			logPrimary.write(b);
		}
	}

	/**
		Writes bytes to Stdout and Primary if needed. Calls write(b,0,b.length)

		@param b Bytes to write to outputstreams
	*/
	public void write( byte b[] ) throws IOException
	{
		//logStdout.println("Using write 2");
		write( b, 0, b.length);
	}

	/**
		Writes bytes to Stdout and Primary if needed. Note: This does use
		the date buffering. 

		@param b Bytes to write to outputstreams
		@param off Offset to start at
		@param len Length of bytes to send to output streams
	*/
	public void write( byte b[], int off, int len) throws IOException
	{
		//logStdout.println("Using write 3");
		if( useStdout == true)
		{
			logDate.setTime(System.currentTimeMillis());
			logStdout.print( logDateFormat.format(logDate) + " : ");
			logStdout.write(b, off, len);
		}
		if( usePrimary == true)
		{
			if( useStdout == false)
				logDate.setTime(System.currentTimeMillis());
			logPrimary.print( logDateFormat.format(logDate) + " : ");
			logPrimary.write(b, off, len);
		}
	}

	/**
		Flushes Buffers of Stdout and Primary
	*/
	public void flush() throws IOException 
	{
		if( useStdout == true)
		{
			logStdout.flush();
		}
		if( usePrimary == true)
		{
			logPrimary.flush();
		}
	}
	/**
		Does not do anything.
	*/
	public void close() throws IOException {}
	
	private void intializeDateHandling()
	{
		logDate = new Date();
		logDateFormat = new SimpleDateFormat("dd MMM yyyy G HH:mm:ss a zzz", Locale.US);
		logDateFormat.setTimeZone(TimeZone.getDefault());

		//logDate.setTime(System.currentTimeMillis());
		//logStream.println( logDateFormat.format(logDate) + " : " + str);
	}
	/**
		Sets whether or not we use date handling

		@param b true means it uses date, false means no
	*/
	public void setuseDate(boolean b)
	{
		useDate = b;
	}
	/**
		Gets whether or not we use date handling
		
		@return true means it uses date, false means no
	*/
	public boolean getuseDate()
	{
		return useDate;
	}
	/**
		Sets the string to use in the internal SimpleDateFormat
		when formatting the date 

		@param str correct string for SimpleDateFormat.applyPattern(str);
	*/
	public void setDateFormat(String str)
	{
		if( str != null)
		{
			logDateFormat.applyPattern(str);
		}
	}
	/**
		Gets the string from the internal SimpleDateFormat
		used when formatting the date 

		@returns correct string for SimpleDateFormat.toPattern();
	*/
	public String getDateFormat()
	{
		return logDateFormat.toPattern();
	}


	/**
		Sets whether or not we print to Stdout

		@param b true means it uses Stdout, false means no
	*/
	public void useLogStdout(boolean b)
	{
		useStdout = b;
	}
	/**
		Gets whether or not we print to Stdout

		@return true means it uses Stdout, false means no
	*/
	public boolean isLoggingStdout()
	{
		return useStdout;
	}
	/**
		Gets Stdout. This class uses System.out

		@return a PrintStream representing System.out
	*/
	public PrintStream getLogStdout()
	{
		return logStdout;
	}

	/**
		Sets whether or not we print to Primary

		@param b true means it uses Primary, false means no
	*/
	public void useLogPrimary(boolean b)
	{
		usePrimary = b;
	}
	/**
		Gets whether or not we print to Primary

		@return true means it uses Primary, false means no
	*/
	public boolean isLoggingPrimary()
	{
		return usePrimary;
	}
	/**
		Sets logPrimary. 

		@param os a PrintStream to represent logPrimary
	*/
	public void setLogPrimary(PrintStream os)
	{
		if( os != null)	
		{
			logPrimary = os;
		}
	}
	/**
		Gets logPrimary. This must be set before or else null will be returned

		@return a PrintStream representing logPrimary
	*/
	public PrintStream getlogPrimary()
	{
		return logPrimary;
	}
}
