/*
    TftpServerConnection.java - a TFTP Protocol Version 2 Server Connection Manager
    Copyright (C) 1999 Mark Benvenuto <mcb54@columbia.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

package gnu.inet.tftp;

//import java.lang.*;
import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.io.IOException;
import java.util.StringTokenizer;

/**
TftpServerConnection

A class to manage either a read or write tftp server side request

@author Mark Benvenuto
@version 0.7
@see Tftpd
@see TftpReadConnection 
@see TftpWriteConnection 

*/
public class TftpServerConnection implements Runnable
{
	InetAddress cia;
	DatagramSocket conn;
	int localport, remoteport;
	String localfile;
	int error, debug, maxPacketSize = 516;
	int TransferType;
	PrintStream logWriter;
	static final short TFTP_RRQ = 1;
	static final short TFTP_WRQ = 2;
	static final short TFTP_DATA = 3;
	static final short TFTP_ACK = 4;
	static final short TFTP_ERROR = 5;
	static final String TFTP_netascii = "netascii";
	static final String TFTP_octet = "octet";
	static final String TFTP_mail = "mail";

	/**
		Creates a new instance of TftpServerConnection 

		@param ia The internet address of the remote host
		@param rport The remote port of the client
		@param localFilename The filename of the local file
		@param rtype The request type, 1 for read, 2 for write
		@param logStream A PrintWriter to logging all errors to
		@param dbstate A flag to enable debug/verbose support, 0 for no debuging
		@see java.net.InternetAddress
	*/
	TftpServerConnection(InetAddress ia, int rport, String localFilename, int rtype, PrintStream logStream, int dstate)
	{
		debug = dstate;
		cia = ia;
		remoteport = rport;
		localfile = localFilename;
		TransferType = rtype;
		logWriter = logStream;

		//Create Socket
		try{
			conn = new DatagramSocket();
		}catch( SocketException e)
		{
			logWriter.println("Datagram Socket Creation Failed: " + e);
			conn = null;
			return;
		}
		
		localport = conn.getLocalPort();

		if( debug > 0)
		{
			logWriter.println("Remote Address: " + cia);
			logWriter.println("Remote Port: " + remoteport);
			logWriter.println("Local File: " + localfile);
			logWriter.println("Transfer Type: " + TransferType);
			logWriter.println("Log Writer: " + logWriter);
		}
	}
	/**
		Processes the TftpServerConnection
	*/
	public void run()
	{
		if(conn == null)
			return;
		if(TransferType == 1)  //Read Request - send file
		{
			//Send the File
			if(debug > 0)
				logWriter.println("Reading the file: " + localfile);

			//Hande Data
			//Check here so that we don't initiate a write even though
			//we might not have the file
			File inFile = new File(localfile);
			if(inFile.exists() == false)
			{
				logWriter.println("File does not exist");
				return;
			}

			if(debug > 0)
				logWriter.println("Created Socket: " + conn);

			try{
				FileInputStream fstream = new FileInputStream(localfile);
				//BufferedInputStream bfstream = new BufferedInputStream(fstream, 512);
				TftpWriteConnection tftpWConn = new TftpWriteConnection(conn, cia, fstream, debug);
				tftpWConn.setLogWriter( logWriter );
				tftpWConn.setConnectionMode(true);
				tftpWConn.setRemotePort( remoteport );
				tftpWConn.processConnection();
				fstream.close();
				logWriter.println("Transfer Successfull Completed");
			}catch(IOException IOe)
			{
				logWriter.println("FileInputStream Creation failed: " + IOe);
				return;
			}
		}
		else if(TransferType == 2) //Write Request - read the file
		{
			//Read the File
			if(debug > 0)
				logWriter.println("Writing the file: " + localfile);

			if(debug > 0)
				logWriter.println("Created Socket: " + conn);

			//Hande Data
			File outFile = new File(localfile);
			try{
				FileOutputStream fstream = new FileOutputStream(localfile);
				TftpReadConnection tftpRConn = new TftpReadConnection(conn, cia, fstream, debug);
				tftpRConn.setLogWriter( logWriter );
				tftpRConn.setConnectionMode(true);
				tftpRConn.setRemotePort( remoteport );
				tftpRConn.processConnection();
				fstream.close();
				logWriter.println("Transfer Successfull Completed");
			}catch(IOException IOe)
			{
				logWriter.println("FileOutputStream Creation failed: " + IOe);
				return;
			}
		}
	}
	/**
		Sets the socket timeout for the connection socket 
		which applies also to TftpReadConnection and TftpWriteConnection.

		@param time The new socket timeout in milliseconds
		@see TftpReadConnection 
		@see TftpWriteConnection 
	*/
	void setSocketTimeout(int time)
	{	
		try{
			conn.setSoTimeout(time);
		}catch( SocketException e)
		{
			System.err.println("Error on setting socket timeout: " + e);
		}
	}	
	/**
		Gets the socket timeout for the connection socket 
		which applies also to TftpReadConnection and TftpWriteConnection.

		@return The socket timeout in milliseconds
		@see TftpReadConnection 
		@see TftpWriteConnection 
	*/
	int getSocketTimeout() throws SocketException
	{	
		return conn.getSoTimeout();
	}	
	/*
		Used to read a string from an arrary of bytes, like a packet
	
		@param rs Bytes from a packet to read for string
		@param off Offset is the position to start reading from
		@return A String from the packet, could be zero length
	*/
	protected String readString(byte[] rs, int off)
	{
		String s = new String();
		off--;
		while(rs[++off]!=(byte)'\0')
			s += String.valueOf((char)rs[off]);
		return s;
	}

}
