/*
    TftpWriteConnection.java - a TFTP Protocol Version 2 Writer
    Copyright (C) 1999 Mark Benvenuto <mcb54@columbia.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
package gnu.inet.tftp;

//import java.lang.*;
import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.util.StringTokenizer;

/**
TftpWriteConnection 

A class to manage an entire write tftp request. 
This class processes the file as a thread.

@author Mark Benvenuto
@version 0.6
@see Tftp
@see TftpReadConnection 
@see TftpConnection 

*/
public class TftpWriteConnection
{
	DatagramSocket sock;
	InetAddress ria;
	FileInputStream istream;
	int remoteport = 0;
	int maxPacketSize = 516;
	int debug=0;
	int maxDATAtries=3;
	boolean connectionMode = false;
	PrintStream logWriter;
	/**
		Creates a new instance of TftpWriteConnection

		@param ds A DatagramSocket for i/o
		@param ia The internet address of the remote host
		@param fstream Stream to read file from
		@param dbstate A flag to enable debug/verbose support, 0 for no debuging
		@see java.net.DatagramSocket
		@see java.net.InternetAddress
	*/
	TftpWriteConnection(DatagramSocket ds, InetAddress ia, FileInputStream fstream, int dbstate)
	{
			sock = ds;
			debug = dbstate;
			ria = ia;
			istream = fstream;
			logWriter = System.out ;
	}
	/**
		Sets PrintStream to use for logging and error reporting

		@param pw PrintStream to log errors to 
	*/
	void setLogWriter( PrintStream pw)
	{
		if( pw != null)
			logWriter = pw;
		else
			logWriter = System.out ;
	}
	/**
		Sets this connection in a server mode. It means that in
		the main loop it will skip the first part of the loop
		when it first goes through the loop

		@param connMode true is server, false is not server (default)
	*/
	void setConnectionMode( boolean connMode)
	{
		connectionMode = connMode;
	}
	/**
		Sets the remote port for this connection in a server mode. 
		It is needed for the server as it must know from where to start
		sending messages to while the client automatically determines this itself.

		@param rport Remote Port to send datagrams too
	*/
	void setRemotePort( int rport)
	{
		if( rport > 0) 
			remoteport = rport;
	}
	/**
		Sets maximum retries for sending DATA, default 3

		@param tries Number of retries
	*/
	void setMaxTries(int tries)
	{
		if(tries > 0)
			maxDATAtries = tries;
	}
	/**
		Sends a DatagramPacket representing an DATA packet

		@param tries Number of retries
		@return true if success, false otherwise
		@see java.net.DatagramPacket
	*/
	private boolean sendDATA(DatagramPacket p)
	{
		try{
			sock.send(p);
			return true;
		}catch(IOException e)
		{
			logWriter.println("Failed to Send DATA packet: " + e);
		}
		return false;
	}
	/**
		Processes the write connection
	*/
	public void processConnection()
	{
		int isFinished = 0;
		int isRead = 1, blocknum = 1, oldblocknum = 0, pblocknum = 0;
		byte[] ACKdata = new byte[maxPacketSize]; //Also used to receive errors
		int DATAtries = 0, RECVtries = 0;
		byte data[] = new byte[maxPacketSize];
		data[0] = 0;
		data[1] = 3;
		int datalength = 516;
		boolean DATAretcode = false;
		DatagramPacket ACKp = new DatagramPacket(ACKdata, ACKdata.length);
		DatagramPacket p = new DatagramPacket(data, maxPacketSize);
		boolean skipLoop = connectionMode;
		if( skipLoop == false)
		{
			remoteport = 0;
		}

		while(isRead == 1)
		{
			if( skipLoop == false)
			{
				if(debug > 0)
					logWriter.println("Receiving Packet");
				try{
					sock.receive(ACKp);
				} 
				catch( InterruptedIOException iioe) {
					logWriter.println("Error Receiving Packet: " + iioe);
					if(RECVtries > maxDATAtries)
					{
						logWriter.println("Waited to many times for ACK packet");
						return;	
					}
					RECVtries++;
					logWriter.println("Continuing to wait for Packet");
					continue;
				}
				catch(IOException e) {
					logWriter.println("Error Receiving Packet: " + e);
					return;
				}
				RECVtries = 0;
				if(remoteport == 0)
				{
					remoteport = ACKp.getPort();
					p.setPort(remoteport);
					p.setAddress(ria);
				}
				datalength = ACKp.getLength();
				if(debug > 0)
				{
					logWriter.println("Received Packet: " + ACKp);
					logWriter.println("Type: " + ACKdata[0] + ACKdata[1]);
					if(ACKdata[1] == 3)
					{
						logWriter.println("Received DATA Packet( Len: " + datalength + " )");
						logWriter.println(ACKp.getData());
					}
					if(ACKdata[1] == 4)
					{
						logWriter.println("Received ACK Packet( Len: " + datalength + " )");
						logWriter.println("ACK: " + (ACKdata[2] << 8) + ACKdata[3]);
					}
				}
				if(isError(ACKdata))
				{
					logWriter.println("Received ERROR Packet");
					printError(data);
					isRead = 0;
					break;
				}
				if(ACKdata[1] != 4)
				{	
					//Failed to get ACK packet
					sendError((short)4, "Expected ACK Packet");
					logWriter.println("Expected ACK Packet");
					break;	
				}
				pblocknum = (ACKdata[2] << 8) + ACKdata[3];
			}
			if( (pblocknum == oldblocknum) || (skipLoop == true))
			{
				if (skipLoop == true)
				{
					oldblocknum = 0;
					pblocknum = 1;
					blocknum = 1;
					skipLoop = false;
					p.setAddress(ria);
					if(remoteport == 0)
					{
						logWriter.println("Failed to set remote port. Exiting");
						return;
					}
					p.setPort(remoteport);
				}
				if(isFinished == 1)
					return;
				//Send next DATA packet
				if(debug > 0)
					logWriter.println("Received Correct Block Num: " + pblocknum);

				//Read DATA from File
				if(debug > 0)
					logWriter.println("Read DATA from File");
				try{
					datalength = istream.read(data, 4, 512);
				}catch(IOException IOe)
				{
					logWriter.println("Failed to read DATA packet from file: " + IOe);
					return;
				}
				if(debug > 0)
					logWriter.println("Read " + datalength + " bytes");
				if(datalength < 512) //Read the last packet
				{
					isFinished = 1;
					if(datalength < 0)
					{
						datalength = 0;
					}
				}
				p.setLength(datalength + 4);

				DATAtries = 0;
				if(debug > 0)
					logWriter.println("Sending the DATA Packet");
				data[2] = (byte)(0xFF & (blocknum >> 8));
				data[3] = (byte)(0xFF & blocknum);
				if(debug > 0)
					logWriter.println("Sending Packet #: " + data[2] + data[3]);
				DATAretcode = sendDATA(p);

				//Exit if sending DATA packet failed here
				if(!DATAretcode)
					return;

				//This is the block we were expecting
				oldblocknum++;
				blocknum++;

				continue;
			}
			if( pblocknum < oldblocknum)
			{
				if(debug > 0)
					logWriter.println("Received Duplicate Block Num: " + pblocknum);
				//Last DATA failed to receive
				if(DATAtries > maxDATAtries)
				{
					//We have retried to many times
					//Send an error message to be nice
					sendError((short)4, "Too Many Retries");

					logWriter.println("Failed to Receive proper ACK for DATA packet");
					if(isFinished == 1)
						logWriter.println("File was sent off completely");
					return;
				}
				//Send it again
				if(debug > 0)
					logWriter.println("Sending the DATA Packet");
				data[3] = (byte)(data[3] + 1);
				if(!sendDATA(p))
					return;
				DATAtries++;
			}
			else if( pblocknum > oldblocknum)
			{
				if(debug > 0)
					logWriter.println("Received Incorrect Block Num: " + pblocknum);
				//We have an invalid block number
				sendError((short)4, "Invalid BLOCK number");
				return;
			}
		}
	}
		/**
		Creates a proper Error packet
		
		@param ErrorCode A code representing the error
		@param str A string describing the error, could be null
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildErrorPacket(short ErrorCode, String str)
	{
		byte data[];
		int pos;
		data = new byte[5 + str.length()];
	

		data[0] = 0;
		data[1] = 5;

		data[2] = (byte)(0xFF & (ErrorCode >> 8));
		data[3] = (byte)(0xFF & ErrorCode);

		//Store String
	        byte bytes[] = str.getBytes();
        	System.arraycopy(bytes, 0, data, 4, bytes.length);
	        data[4 + bytes.length] = 0;
		
		return new DatagramPacket(data, data.length, ria, remoteport);
	}
	/**
		Send remote computer error packet using <B>BuildErrorPacket</B>
		
		@param ErrorCode A code representing the error
		@param str A string describing the error, could be null
	*/
	void sendError(short ErrorCode, String str)
	{
		try{
			sock.send(BuildErrorPacket(ErrorCode, str));
		}catch( IOException e)
		{
			logWriter.println("Error on Sending ERROR Packet: " + e);
		}
	}
	/**
		Checks if packet is error packet
		
		@param data Bytes from a packet to check if it is an error packet. Minimum length 2
		@return True if error packet, false otherwise
	*/
	boolean isError(byte[] data)
	{
		if(data[1] == 5)
			return true;
		return false;
	}
	/**
		Prints error packet. If there is not string in the packet, it uses a standard string defined by the RFC.
		
		@param data Bytes from an error packet
	*/
	void printError(byte[] data)
	{
		logWriter.println("Error: " + (short)((data[2] << 8) + data[3]));
		String str = readString(data, 4);
		if(str.length() == 0)
			logWriter.println("ErrorStr: " + str);
		else
		{
			switch(data[3])
			{
				case 0:
					logWriter.println("Not defined, see error message (if any).");
					break;
				case 1:
					logWriter.println("File not found.");
					break;
				case 2:
					logWriter.println("Access violation.");
					break;
				case 3:
					logWriter.println("Disk full or allocation exceeded.");
					break;
				case 4:
					logWriter.println("Illegal TFTP operation.");
					break;
				case 5:
					logWriter.println("Unknown transfer ID.");
					break;
				case 6:
					logWriter.println("File already exists.");
					break;
				case 7:
					logWriter.println("No such user.");
					break;
				default:
					logWriter.println("File not found.");
					break;
			}
		}
	}
	/*
		Used to read a string from an arrary of bytes, like a packet
	
		@param rs Bytes from a packet to read for string
		@param off Offset is the position to start reading from
		@return A String from the packet, could be zero length
	*/
	protected String readString(byte[] rs, int off)
	{
		String s = new String();
		off--;
		while(rs[++off]!=(byte)'\0')
			s += String.valueOf((char)rs[off]);
		return s;
	}
}
